    # item location format:
    # (code, (qch, qch))
    # 0 - floor, location = (x,y,t/f)
    # 1 - inven, location = (actorID)
    # 2 - equip, location = (actorID, slot)
    # 3 - nowhere

import random
import kdata
import curses
from math import sqrt
from string import ascii_uppercase

class kitem(object):
  def __init__(self, char, name):
    self.char = char
    self.orig_name = name
    self.name = name
    self.inscription = ''
    self.equippable = False
    self.identified = False
    self.value = 0
    self.colour = curses.COLOR_BLUE
  def update_name(self):
    if self.inscription != '':
      self.name = '{'+self.inscription+'} '+self.orig_name
    else:
      self.name = self.orig_name

class kwall(kitem):
  def __init__(self, char):
    super(kwall, self).__init__(char, 'a piece of wall')

class kequip(kitem):
  def __init__(self, char, slot, name, stats, resistances, brands):
    super(kequip,self).__init__(char, name)
    self.equippable = True
    self.slot = slot
    self.stats = stats
    self.resistances = set(resistances)
    self.brands = set(brands)
    self.colour = curses.COLOR_GREEN

  def enchant(self):
    if self.char in ['/','|']:
      self.brands.update([random.choice(['pois','fire','cold','elec','holy',
                                      'deth','tele','vamp','chao','thef'])])
      if self.char == '/':
        self.brands.update([random.choice(['pois','fire','cold','elec',
                             'holy','deth','tele','vamp','chao','thef'])])


    else:
      self.resistances.update([random.choice(['pois','fire','cold',
                      'elec','holy','deth','tele','vamp','chao','thef'])])

    for stat in enchant_stats[self.char]:
      r = random.randint(0,1+self.stats[stat]/2)
      if stat == 'hp':
        r *= 3
      if self.char == '/':
        r *= 2
      if random.random() < 0.5:
        self.stats[stat] += r
    self.orig_name = '*'+self.orig_name
    self.update_name()

  def damage(self):
    if self.char in ['/','|'] and len(self.brands) > 0:
      self.brands.pop()
    elif len(self.resistances) > 0:
      self.resistances.pop()
    for stat,val in self.stats.items():
      if random.random() < 0.5 and stat != 'dice':
        self.stats[stat] /= 2
    self.orig_name = '-'+self.orig_name
    self.update_name()
      
    

class knull_item(kequip):
  def __init__(self):
    stats = {'spd':0, 'hp':0, 'atk':0,'def':0, 'dice':(0,0),
             'dam':0,'arm':0, 'cat':0,'cdf':0}
    super(knull_item ,self).__init__('_','','nothing',
                                     stats, set(), set())

potion_flavours = []
def random_pot_name():
  for i in range(20):
    p = (random.choice(kdata.potion_textures) + ' ' +
         random.choice(kdata.potion_colours))
    if p not in potion_flavours:
      potion_flavours.append(p)
      return p
  raise Exception('death')

scroll_titles = []
def random_scroll_title():
  for i in range(20):
    t = ''.join([random.choice(ascii_uppercase)
                 for j in range(random.randint(5,10))])
    if t not in scroll_titles:
      scroll_titles.append(t)
      return t
  raise Exception('death')

enchant_stats = {}
enchant_stats['|'] = ['atk','dam','cat','spd','hp','def']
enchant_stats['/'] = ['atk','dam','cat','spd','hp','def']

def random_weapon():
  name,char,quality,n,d = random.choice(kdata.weapon_types)
  if random.choice([True,False]):
    quality = random.randint(1,quality)
  if random.choice([True,False]):
    quality += 1
  name = 'the '+name+' of '+random.choice(
    kdata.weapon_flavours[quality-1:quality+4])
  stats = dict(zip(['atk','dam','cat','spd','hp','def'],
                   reversed(sorted([random.randint(0,quality)
                                    for i in range(6)]))))
  stats['dice'] = (n,d)
  stats['arm'] = 0
  stats['cdf'] = 0
  stats['hp'] *= 3
  brands = set()
  for i in range(quality):
    if random.random() < 0.6/(len(brands)+1):
      brands.update([random.choice(['pois','fire','cold','elec','holy',
                                      'deth','tele','vamp','chao','thef'])])
  eq= kequip(char,'weapon',name,stats,set(),brands)
  eq.value = rate_equipment(eq)
  return eq

enchant_stats[']'] = ['arm','def','hp','cdf']

def random_armour():
  name,char,quality = random.choice(kdata.armour_types)
  if random.choice([True,False]):
    quality = random.randint(1,quality)
  if random.choice([True,False]):
    quality += 1
  name = name + ' of ' + random.choice(
    kdata.armour_flavours[quality-1:quality+2])
  stats = dict(zip(['arm','def','hp','cdf'],reversed(sorted([
    random.randint(0,quality) for i in range(4)]))))
  stats['dice'] = (0,0)
  stats['atk'] = 0
  stats['dam'] = 0
  stats['cat'] = 0
  stats['spd'] = 0
  stats['hp'] *= 10
  resistances = set()
  for i in range(quality):
    if random.random() < 0.6/(len(resistances)+1):
      resistances.update([random.choice(['pois','fire','cold','elec',
                                      'holy','deth','tele','vamp','conf'])])
  if random.random() < 0.01:
    resistances.update(['phys'])
  eq= kequip(char,'armour',name,stats,resistances,set())
  eq.value = rate_equipment(eq)
  return eq

enchant_stats['{'] = ['spd','def','arm','hp','cdf','atk']

def random_boots():
  name,char,quality = random.choice(kdata.boot_types)
  if random.choice([True,False]):
    quality = random.randint(1,quality)
  if random.choice([True,False]):
    quality += 1
  name = 'a pair of '+ name + ' of ' + random.choice(
    kdata.boot_flavours[max(0,2*quality-3):2*quality+1])
  stats = dict(zip(['spd','def','arm','hp','cdf','atk'],reversed(sorted([
    random.randint(0,quality) for i in range(6)]))))
  stats['dice'] = (0,0)
  stats['dam'] = 0
  stats['cat'] = 0
  stats['hp'] *= 3
  stats['spd'] *= 3
  resistances = set()
  for i in range(quality):
    if random.random() < 0.3/(len(resistances)+1):
      resistances.update([random.choice(['pois','fire','cold','elec',
                                      'holy','deth','tele','vamp','conf'])])
  eq= kequip(char,'feet',name,stats,resistances,set())
  eq.value = rate_equipment(eq)
  return eq

enchant_stats['('] = ['def','hp','cdf','arm']

def random_helmet():
  name,char,quality = random.choice(kdata.helmet_types)
  if random.choice([True,False]):
    quality = random.randint(1,quality)
  if random.choice([True,False]):
    quality += 1
  name = 'a '+name + ' of ' + random.choice(
    kdata.helmet_flavours[quality-1:quality+3])
  stats = dict(zip(['def','hp','cdf','arm'],reversed(sorted([
    random.randint(0,quality) for i in range(4)]))))
  stats['dice'] = (0,0)
  stats['atk'] = 0
  stats['dam'] = 0
  stats['cat'] = 0
  stats['spd'] = 0
  stats['cdf'] *= 2
  stats['hp'] *= 5
  resistances = set()
  for i in range(quality):
    if random.random() < 0.6/(len(resistances)+1):
      resistances.update([random.choice(['pois','fire','cold','elec',
                                      'holy','deth','tele','vamp','conf'])])
  eq= kequip(char,'head',name,stats,resistances,set())
  eq.value = rate_equipment(eq)
  return eq

enchant_stats[')'] = ['def','hp','arm','cdf']

def random_shield():
  name,char,quality = random.choice(kdata.shield_types)
  if random.choice([True,False]):
    quality = random.randint(1,quality)
  if random.choice([True,False]):
    quality += 1
  name = 'a '+name + ' of ' + random.choice(
    kdata.shield_flavours[quality-1:quality+3])
  stats = dict(zip(['def','hp','arm','cdf'],reversed(sorted([
    random.randint(0,quality) for i in range(4)]))))
  stats['dice'] = (0,0)
  stats['atk'] = 0
  stats['dam'] = 0
  stats['cat'] = 0
  stats['spd'] = 0
  stats['def'] *= 2
  stats['hp'] *= 5
  resistances = set()
  for i in range(quality):
    if random.random() < 0.8/(len(resistances)+1):
      resistances.update([random.choice(['pois','fire','cold','elec',
                                      'holy','deth','tele','vamp','conf'])])
  eq= kequip(char,'shield',name,stats,resistances,set())
  eq.value = rate_equipment(eq)
  return eq

enchant_stats['}'] = ['def','atk','arm','cdf']

def random_gloves():
  name,char,quality = random.choice(kdata.glove_types)
  if random.choice([True,False]):
    quality = random.randint(1,quality)
  if random.choice([True,False]):
    quality += 1
  name = 'a pair of '+name + ' of ' + random.choice(
    kdata.glove_flavours[quality-1:quality+4])
  stats = dict(zip(['def','atk','arm','cdf'],reversed(sorted([
    random.randint(0,quality) for i in range(4)]))))
  stats['dice'] = (0,0)
  stats['dam'] = 0
  stats['cat'] = 0
  stats['spd'] = 0
  stats['def'] *= 2
  stats['hp'] = 0
  resistances = set()
  for i in range(quality):
    if random.random() < 0.4/(len(resistances)+1):
      resistances.update([random.choice(['pois','fire','cold','elec',
                                      'holy','deth','tele','vamp','conf'])])
  eq= kequip(char,'gloves',name,stats,resistances,set())
  eq.value = rate_equipment(eq)
  return eq

enchant_stats['='] = ['atk','def','arm','dam','cat','cdf','spd','hp']

def random_ring():
  name = random.choice(kdata.ring_types)
  quality = random.randint(5,20)
  name = 'a '+name+' ring of '+random.choice(kdata.ring_flavours)
  stats = dict(zip(['atk','def','arm','dam','cat','cdf','spd','hp'],
                   [random.randint(0,quality) for i in range(8)]))
  stats['dice'] = (0,0)
  stats['hp'] *= random.randint(1,10)
  for stat in ['atk','def','arm','dam','cat','cdf','spd','hp']:
    if random.choice([False,False,True]):
      stats[stat] = 0
  res_or_brands = set()
  for i in range(quality*2):
    if random.random() < 0.3/(len(res_or_brands)+1):
      res_or_brands.update([random.choice(['pois','fire','cold','elec',
                                      'holy','deth','tele','vamp'])])
  eq= kequip('=','ring',name,stats,set(),res_or_brands)
  eq.value = rate_equipment(eq)
  return eq

enchant_stats['"'] = ['atk','def','arm','dam','cat','cdf','spd','hp']              

def random_amulet():
  name = random.choice(kdata.ring_types)
  quality = random.randint(5,20)
  name = 'a '+name+' amulet of '+random.choice(kdata.ring_flavours)
  stats = dict(zip(['atk','def','arm','dam','cat','cdf','spd','hp'],
                   [random.randint(0,quality) for i in range(8)]))
  stats['dice'] = (0,0)
  stats['hp'] *= random.randint(1,10)
  for stat in ['atk','def','arm','dam','cat','cdf','spd','hp']:
    if random.choice([False,False,True]):
      stats[stat] = 0
  res_or_brands = set()
  for i in range(quality*2):
    if random.random() < 0.3/(len(res_or_brands)+1):
      res_or_brands.update([random.choice(['pois','fire','cold','elec',
                                      'holy','deth','tele','vamp'])])
  eq= kequip('"','amulet',name,stats,res_or_brands,set())
  eq.value = rate_equipment(eq)
  return eq

def cool_weapon():
  stats = {'spd':15, 'hp':10, 'atk':5,'def':5, 'dice':(2,8),
             'dam':2,'arm':3, 'cat':10,'cdf':0}
  brands = set(['pois','cold'])
  eq= kequip('|','weapon','The Sword of Justice',stats,set(),brands)
  eq.value = rate_equipment(eq)
  return eq

def sweet_weapon():
  stats = {'spd':15, 'hp':10, 'atk':5,'def':5, 'dice':(3,6),
             'dam':2,'arm':3, 'cat':10,'cdf':0}
  brands = set(['deth','holy','tele'])
  eq= kequip('|','weapon','The Sword of Truth',stats,set(),brands)
  eq.value = rate_equipment(eq)
  return eq

def rate_brands(brands):
  r = 1.0
  for b in brands:
    r *= 1.5
    if b in ['chao','conf','thef']:
      r *= 1.5
  return r

def rate_equipment(item):
  if isinstance(item,knull_item):
    return 0
  if not item.equippable:
    return item.value
  stats = item.stats
  dice = stats['dice']
  av_damage = dice[0]*0.5*(dice[1]+1)+stats['dam']
  return sqrt(stats['spd']+4)*(av_damage+1)*(stats['atk']+stats['def']+5) * (
    1+stats['cat']+stats['cdf'])*(stats['arm']+2)*(sqrt(stats['hp']+20))*(
      1+(av_damage*rate_brands(item.brands)))

def broken_pot():
  return kitem('~','a broken pot')

class kpotion(kitem):
  def __init__(self, function,value):
    super(kpotion,self).__init__('!','a '+random_pot_name()+' potion')
    self.function = function
    self.value = value
    
  def identify(self):
    if not self.identified:
      self.orig_name = self.orig_name+' of '+(self.function.__name__[4:])
    self.identified = True
    self.update_name()

class kscroll(kitem):
  def __init__(self, function,value):
    super(kscroll,self).__init__('?','a scroll titled '+random_scroll_title())
    self.function = function
    self.value = value
    
  def identify(self):
    if not self.identified:
      self.name = self.name+' of '+(self.function.__name__[4:])
    self.identified = True

class InventoryException(Exception):
  pass
  
class kinventory:
  def __init__(self,size):
    self.inv = [0]*size
    self.size = size

  def add(self, itemID):
    for i in range(self.size):
      if self.inv[i] == 0:
        self.inv[i] = itemID
        return True
    raise InventoryException("tried to put an item in a full inventory")

  def drop(self,itemID):
    for i in range(self.size):
      if self.inv[i] == itemID:
        self.inv[i] = 0
        return True
    raise InventoryException("tried to remove an item not in inventory")

  def __contains__(self, itemID):
    return itemID in self.inv

  def __getitem__(self, index):
    return self.inv[index]

  def full(self):
    return not 0 in self.inv

  def index(self, itemID):
    return self.inv.index(itemID)

  def empty(self):
    for i in self.inv:
      if i != 0:
        return False
    return True

  def top(self):
    for i in reversed(self.inv):
      if i != 0:
        return i
    return 0
##    for i in range(1,self.size+1):
##      if self.inv[-i] != 0:
##        return self.inv[-i]
##    return 0
  
  def iteritems(self):
    return zip(range(self.size),self.inv)

slots = {'shield':2,'weapon':0,'armour':1,'head':3, 'feet':4,
         'gloves':5,'amulet':6, 'ring':7, 'unused1':8,'unused2':9}
indices = dict((v,k) for k, v in slots.iteritems())
class kequipment:
  def __init__(self):
    self.eq = {'shield':0,'weapon':0,'armour':0,'head':0, 'feet':0,
               'gloves':0,'ring':0, 'amulet':0, 'unused1':0,'unused2':0}
  
  def __getitem__(self,index):
    if index in slots:
      return self.eq[index]
    if index in indices:
      return self.eq[indices[index]]
    raise KeyError(str(index)+' not in kequipment')

  def __contains__(self,itemID):
    return itemID in self.eq.itervalues()

  def drop(self,itemID):
    for slot,item in self.eq.iteritems():
      if item == itemID:
        self.eq[slot] = 0

  def __setitem__(self,index,value):
    if index in slots:
      self.eq[index]=value
    elif index in indices:
      self.eq[indices[index]]=value
  def iteritems(self):
    return [(slots[idx],itemID) for idx,itemID in self.eq.iteritems()]
  def itervalues(self):
    return self.eq.itervalues()
  
