#!/usr/bin/python
try:
  import curses
except ImportError:
  print "Couldn't import a curses module, please install one."
import random
import krltools as k
import kmap
import kactors
import ktime
import kitems
import kdata
from math import sqrt, log
width, height = 23, 23
apx, apy = 15,15
k.setsize(width,height) #width, height
k.setorientation(random.choice([True,False]))
fov_tree = k.all_bres(min(apx,apy)/2)
debug = False
no_fov = False

class krl:
  def __init__(self, screen):
    self.screen = screen
    self.grid = kmap.random_maze(width, height)
    self.population = {0:None}
    self.items = {0:kitems.knull_item(),
                  -1:kitems.kwall('#'),
                  -2:kitems.kwall('&'),
                  -3:kitems.broken_pot()}
    self._next_actorID = 1
    self._next_itemID = 1
    self.ticker = ktime.ticker()
    self.scent_maps = {}
    self.daemons = {}
    self.message_length = 55
    self.message_number = 20
    self.empty_message = ' '*self.message_length
    self.messages = [self.empty_message]*self.message_number
    self.abyss = kitems.kinventory(20)
    self.potions = []
    self.scrolls = []
    self.uses = {}

  def get_new_actorID(self):
    self._next_actorID += 1
    return self._next_actorID - 1

  def get_new_itemID(self):
    self._next_itemID += 1
    return self._next_itemID - 1

  def add_actor(self, actor, location):
    actorID = self.get_new_actorID()
    if self.population.has_key(actorID):
      raise IDException("tried to use an actor ID that's already taken")
    self.population[actorID] = actor
    actor.ID = actorID
    tile = k.mget(self.grid, location)
    if tile.blocked():
      raise TileBlockedException("tried to place actor on blocked tile")
    actor.location = location
    tile.occupant = actorID
    return actorID

  def add_daemon(self, daemon):
    actorID = self.get_new_actorID()
    self.population[actorID] = daemon
    daemon.ID = actorID
    return actorID

  def add_item(self, item, (loctype, location)):
    itemID = self.get_new_itemID()
    if self.items.has_key(itemID):
      raise IDException("tried to use an item ID that's already taken")
    self.items[itemID] = item
    item.ID = itemID
    if not self.try_place_item(itemID, (loctype, location)):
      raise Exception("couldn't add item")
    return itemID
    #item.location = (loctype, location)
    #unnecessary, prevents nice pots etc.
    
    # item location format:
    # (code, (qch, qch))
    # 0 - floor, location = (x,y,t/f)
    # 1 - inven, location = (actorID)
    # 2 - equip, location = (actorID, slot)
    # 3 - nowhere

  def try_place_item(self,itemID, (loctype, loc)):
    #try to place something, make no changes if placement is impossible
    if loctype == 3:
      return True
    elif loctype == 0:
      tile = k.mget(self.grid, loc)
      if not tile.can_contain():
        return False
      tile.items.add(itemID)
      return True
    elif loctype == 1:
      actor = self.population[loc]
      if actor.inventory.full():
        return False
      actor.inventory.add(itemID)
      return True
    elif loctype == 2:
      actor = self.population[loc[0]]
      if not self.items[itemID].equippable:
        return False
      slot = self.items[itemID].slot
      if slot != loc[1]:
        return False
      else:
        #replace = actor.equip[loc[1]]
        if actor.equip[loc[1]] in [0,itemID]:
          actor.equip[loc[1]] = itemID
          return True
        else:
          actor.equip[loc[1]] = itemID
          return True

  def del_actor(self, actorID):
    k.mget(self.grid, self.population[actorID].location).occupant = 0
    del self.population[actorID]


##  def del_item(self, itemID):
##    loctype, location = self.items[itemID].location
##    if loctype == 0:
##      k.mget(self.grid, location).items.drop(itemID)
##    elif loctype == 1:
##      self.population[location[0]].inventory.drop(itemID)
##    elif loctype == 2:
##      self.population[location[0]].equip[location[1]] = 0
##    del self.items[itemID]

  def resist(self, actorID, element):
    if 'resi' in self.population[actorID].status:
      if element in ['pois','fire','cold','elec','holy','deth','conf','tele']:
        return True
    for equipID in self.population[actorID].equip.itervalues():
      if element in self.items[equipID].resistances:
        return True
    return False

  def brands(self, actorID):
    b = set(['phys'])
    for equipID in self.population[actorID].equip.itervalues():
      b.update(self.items[equipID].brands)
    return b

  def damage_roll(self, itemID):
    n,d = self.items[itemID].stats['dice']
    dam = sum([random.randint(1,d) for i in range(n)])
    #self.message(str(dam)+' damage')
    return sum([random.randint(1,d) for i in range(n)])

  def calc_stat(self, actorID, stat):
    if actorID in self.daemons.values():
      return 50
    actor = self.population[actorID]
    value = actor.base_stats[stat]
    if stat == 'hp':
      div = 1
      for itemID in actor.equip.itervalues():
        value = value*(100+self.items[itemID].stats[stat])
        div *= 100
      return value/div
    value += sum([self.items[itemID].stats[stat]
                                 for itemID in actor.equip.itervalues()])
    if 'berz' in actor.status:
      value += {'spd':2, 'hp':0, 'atk':10,'def':-5,
                'dam':15,'arm':1, 'cat':5,'cdf':0}[stat]
    if stat == 'spd':
      if 'slow' in actor.status:
        if 'fast' not in actor.status:
          return value*5/8
      elif 'fast' in actor.status:
        return value*8/5
    return value

  def calc_damage(self, actorID):#(damage roll, bonus damage stat)
    actor = self.population[actorID]
    return (self.damage_roll(actor.equip['weapon']),
              actor.base_stats['dam'] + sum([self.items[itemID].stats['dam']
                 for itemID in actor.equip.itervalues()]))

  def heal(self,actorID,hp):
    actor = self.population[actorID]
    actor.hp = min(self.calc_stat(actorID,'hp'),actor.hp + hp)
  
  def temp_resi(self,actorID,magic,attackerID):
    if actorID not in self.population:
      return 0
    if magic > 1:
      self.event_message(actorID,0,'res2')
      self.population[actorID].status['resi'] = True
      return min(magic-1,100)
    del self.population[actorID].status['resi']
    self.event_message(actorID,0,'res3')
    return 0
  
  def berserk(self,actorID,magic,attackerID):
    if actorID not in self.population:
      return 0
    if magic > 1:
      self.event_message(actorID,0,'berz2')
      self.population[actorID].status['berz'] = True
      return min(magic-1,100)
    del self.population[actorID].status['berz']
    self.event_message(actorID,0,'berz3')

  def poison_damage(self, actorID, magic, attackerID):
    if actorID not in self.population:
      return 0
    dam = 1+int(log(abs(1+magic),2)/abs(self.calc_stat(actorID,'spd')))
    if magic-dam < 1:
      del self.population[actorID].status['pois']
      self.event_message(actorID,0,'pois3')
    elif random.choice([True,False]):
      if not self.resist(actorID, 'pois'):
        #self.message(self.population[actorID].name +
        #             ' looks ill. ('+str(dam)+')')
        self.event_message(actorID,0,'pois2')
        if self.deal_damage(actorID,attackerID,dam,0,['magi']):
          self.actor_kill(actorID, attackerID)
    return magic - dam
    
  def fire_damage(self, actorID, magic, attackerID):
    if actorID not in self.population:
      return 0
    dam = 1+(magic*1/20)
    #self.message(self.population[actorID].name+' is on fire! ('+str(dam)+')')
    self.event_message(actorID,0,'fire2')
    if magic-dam < 1:
      del self.population[actorID].status['fire']
      self.event_message(actorID,0,'fire3')
    elif not self.resist(actorID, 'fire'):
      if self.deal_damage(actorID,attackerID,dam,0,['magi']):
        self.actor_kill(actorID, attackerID)
    return magic - dam
    
  def slow(self, actorID, magic, attackerID):
    if actorID not in self.population:
      return 0
    if magic > 1:
      if not self.resist(actorID, 'cold'):
        #self.message(self.population[actorID].name+' struggles to move.')
        self.event_message(actorID,0,'cold2')
        self.population[actorID].status['slow'] = True
      return min(magic-1,100)
    del self.population[actorID].status['slow']
    self.event_message(actorID,0,'cold3')
    return min(magic-1,100)

  def fast(self, actorID,magic, attackerID):
    if actorID not in self.population:
      return 0
    if magic > 1:
      self.population[actorID].status['fast'] = True
      self.event_message(actorID,0,'fast2')
      #self.message(self.population[actorID].name+' is a blur!.')
      return min(magic-1,100)
    del self.population[actorID].status['fast']
    self.event_message(actorID,0,'fast3')
    return min(magic-1,100)

  def conf(self, actorID, magic, attackerID):#todo implement conf!
    if actorID not in self.population:
      return 0
    if magic > 1:
      if not self.resist(actorID, 'conf'):
        self.message(self.population[actorID].name+' stumbles.')
        self.population[actorID].status['conf'] = True
      return min(magic-1,100)
    del self.population[actorID].status['conf']
    self.event_message(actorID,0,'conf3')
    return min(magic-1,100)
  
  def inflict(self, actorID, element, magic,attackerID):
    actor = self.population[actorID]
    actor.status[element] = True
    try:
      self.population[self.daemons[element]].register[actorID] += magic
    except KeyError:
      self.population[self.daemons[element]].register[actorID] = magic
    self.population[self.daemons[element]].causes[actorID] = attackerID

  def actor_visible(self,actorID):
    return (k.mget(self.grid,self.population[actorID].location).last_seen
            - self.ticker.ticks > -self.calc_stat(1,'spd'))

  def event_message(self,actor1,actor2,mtype):
    p1 = actor1 == 1
    if self.actor_visible(actor1):
      name1 = self.population[actor1].name
    else:
      name1 = 'something'
    if isinstance(mtype,int) or isinstance(mtype,long):
      if actor2 == 0:
        if self.actor_visible(actor1):
          if p1:
            self.message('you have '+str(mtype)+' hp.')
          else:
            self.message(name1+' has '+str(mtype)+' hp.')
        return
    if mtype == 'pois1':
      if p1:
        self.message('you are poisoned!')
      elif self.actor_visible(actor1):
        self.message(name1+' is poisoned!')
      return
    elif mtype == 'pois2':
      if p1:
        self.message('you feel ill.',False)
      elif self.actor_visible(actor1):
        self.message(name1+' looks ill.',False)
      return
    elif mtype == 'pois3':
      if p1:
        self.message('you feel less sick.')
      elif self.actor_visible(actor1):
        self.message(name1+' looks less sick.')
      return
    elif mtype == 'fire1':
      if p1:
        self.message('you are set alight!')
      elif self.actor_visible(actor1):
        self.message(name1+' is set alight!')
      return
    elif mtype == 'fire2':
      if p1:
        self.message('you are on fire!',False)
      elif self.actor_visible(actor1):
        self.message(name1+' is on fire!',False)
      return
    elif mtype == 'fire3':
      if p1:
        self.message('you stop burning.')
      elif self.actor_visible(actor1):
        self.message(name1+' stops burning.')
      return
    elif mtype == 'cold1':
      if p1:
        self.message('you are frozen!')
      elif self.actor_visible(actor1):
        self.message(name1+' is frozen!')
      return
    elif mtype == 'cold2':
      if p1:
        self.message('you struggle to move.',False)
      elif self.actor_visible(actor1):
        self.message(name1+' struggles to move.',False)
      return
    elif mtype == 'cold3':
      if p1:
        self.message('you speed up again.')
      elif self.actor_visible(actor1):
        self.message(name1+' speeds up again.')
      return
    elif mtype == 'res1':
      if p1:
        self.message('you feel resistant to the elements!')
      elif self.actor_visible(actor1):
        self.message(name1+' looks resistant to the elements!')
      return
    elif mtype == 'res2':
      if p1:
        self.message('you are resistant.',False)
      elif self.actor_visible(actor1):
        self.message(name1+' looks resistant.',False)
      return
    elif mtype == 'res3':
      if p1:
        self.message('you feel less resistant.')
      elif self.actor_visible(actor1):
        self.message(name1+' looks less resistant.')
      return
    elif mtype == 'berz1':
      if p1:
        self.message('you go berserk!')
      elif self.actor_visible(actor1):
        self.message(name1+' goes berserk!')
      return
    elif mtype == 'berz2':
      if p1:
        self.message('you are very angry!',False)
      elif self.actor_visible(actor1):
        self.message(name1+' looks very angry!',False)
      return
    elif mtype == 'berz3':
      if p1:
        self.message('you feel less angry.')
      elif self.actor_visible(actor1):
        self.message(name1+' looks less angry.')
      return
    elif mtype == 'skill':
      if p1:
        self.message('you feel very skillful!')
      elif self.actor_visible(actor1):
        self.message(name1+' looks more confident.')
      return
    elif mtype == 'elec':
      if p1:
        self.message('you are struck by lightning!')
      elif self.actor_visible(actor1):
        self.message('lightning strikes '+name1+'!')
      return
    elif mtype == 'holy':
      if self.actor_visible(actor1):
        self.message('there is a flash of light!')
      return
    elif mtype == 'deth':
      if self.actor_visible(actor1):
        self.message('everything goes dark for a moment...')
      return
    elif mtype == 'tele1':
      if p1:
        self.message('you disappear!')
      elif self.actor_visible(actor1):
        self.message(name1+' disappears!')
      return
    elif mtype == 'tele2':
      if p1:
        self.message('you reappear!')
      elif self.actor_visible(actor1):
        self.message(name1+' appears!')
      return
    elif mtype == 'tele3':
      if p1:
        self.message('you fade for a moment.')
      elif self.actor_visible(actor1):
        self.message(name1+' fades for a moemnt.')
      return
    elif mtype == 'thef':
      if self.actor_visible(actor1):
        self.message(name1+' stole '+self.items[actor2].name+'!')
      return
    elif mtype == 'vamp':
      if p1:
        self.message('you feel better.')
      elif self.actor_visible(actor1):
        self.message(name1+' looks healthier.')
      return
    elif mtype == 'die':
      if p1:
        self.message('You die.')
      elif self.actor_visible(actor1):
        self.message(name1+' dies.')
      return
    elif mtype == 'fast1':
      if p1:
        self.message('you start running!')
      elif self.actor_visible(actor1):
        self.message(name1+' starts running!')
      return
    elif mtype == 'fast2':
      if p1:
        self.message('you are a blur!',False)
      elif self.actor_visible(actor1):
        self.message(name1+' is a blur!',False)
      return
    elif mtype == 'fast3':
      if p1:
        self.message('you slow down again.')
      elif self.actor_visible(actor1):
        self.message(name1+' slows down again.')
      return
    elif mtype == 'equip':
      if not p1:
        if self.actor_visible(actor1):
          self.message(name1+' equips something.')
      return
      
    p2 = actor2 == 1
    if actor2 == 0:
      name2 = 'WHUT'
      self.message(' '.join([str(actor1),str(actor2),str(mtype)])+
                   ' is a miserable failure.')
      self.message(str(type(mtype)))
      return
    elif self.actor_visible(actor2):
      name2 = self.population[actor2].name
    else:
      name2 = 'something'
    if self.actor_visible(actor2) or self.actor_visible(actor1):
      if mtype == 'hit':
        weapon = self.items[self.population[actor1].equip['weapon']].name
        if weapon == 'nothing':
          if p1:
            if name2 == 'La Femme Fatale':
              self.message('vous donnez un coup de poing a '+name2+'.')
            else:
              self.message('you punch '+name2+'.')
          else:
            if name2 == 'La Femme Fatale' or name1 == 'La Femme Fatale':
              self.message(name1+' donne un coup de poing a '+name2+'.')
            else:
              self.message(name1+' punches '+name2+'.')
          return
        if p1:
          if name2 == 'La Femme Fatale':
            self.message('vous frappez '+name2+' avec '+weapon+'.')
          else:
            self.message('you hit '+name2+' with '+weapon+'.')
          self.increment_uses(weapon)
        else:
          if name1 == 'La Femme Fatale' or name2 == 'La Femme Fatale':
            if actor2 == 1:
              self.message(name1+' vous frappe avec '+weapon+'.')
            else:
              self.message(name1+' frappe '+name2+' avec '+weapon+'.')
          else:
            self.message(name1+' hits '+name2+' with '+weapon+'.')
        return
      elif mtype == 'miss':
        if p1:
          self.message('you miss '+name2+'.')
        else:
          self.message(name1+' misses '+name2+'.')
        return
      elif isinstance(mtype,int) or isinstance(mtype,long):
        if mtype == 2:
          self.message('It was a good hit!')
        elif mtype == 2:
          self.message('It was a great hit!')
        else:
          self.message('It was a '+'*'*(mtype-3)+
                       'SUPERB'+'*'*(mtype-3)+' hit'+'!'*(mtype-2))
        

  def deal_damage(self, actorID, attackerID, damage, magic, dam_types):
    actor = self.population[actorID]#who is hurt
    #self.message(str(dam_types))
    do_damage = False
    if 'pois' in dam_types:
      if not self.resist(actorID, 'pois'):
        do_damage = True
        self.inflict(actorID, 'pois', min(20,magic*2),attackerID)
        self.event_message(actorID,0,'pois1')
        #self.message(actor.name+' is poisoned!')
    if 'fire' in dam_types:
      if not self.resist(actorID, 'fire'):
        do_damage = True
        self.inflict(actorID, 'fire', magic*2,attackerID)
        self.event_message(actorID,0,'fire1')
        #self.message(actor.name+' is set alight!')
    if 'cold' in dam_types:
      if not self.resist(actorID, 'cold'):
        do_damage = True
        self.inflict(actorID, 'slow', min(30,magic*2),attackerID)
        self.event_message(actorID,0,'cold1')
        #self.message(actor.name+' is frozen!')
    if 'elec' in dam_types:
      if not self.resist(actorID, 'elec'):
        do_damage = True
        if random.random() < 1.0/3:
          actor.hp -= random.randint(1,1+(damage/3))
          self.event_message(actorID,0,'elec')
          #self.message('lightning strikes '+actor.name+'!')
    if 'holy' in dam_types:
      if not self.resist(actorID, 'holy'):
        do_damage = True
        holy_dam = random.randint(0,self.population[attackerID].pray_count)
        if holy_dam > 0:
          self.event_message(actorID,0,'holy')
          #self.message('There is a flash of light!')
          actor.hp -= holy_dam
    if 'deth' in dam_types:
      if not self.resist(actorID,'deth'):
        do_damage = True
        death_dam = random.randint(0,self.population[attackerID].kill_count)
        if death_dam > 0:
          self.event_message(actorID,0,'deth')
          #self.message('everything goes dark for a moment...')
          actor.hp -= death_dam

    if 'conf' in dam_types:#not implemented
      if not self.resist(actorID,'conf'):
        do_damage = True
        #self.inflict(actorID, 'conf', magic)
    if 'chao' in dam_types:
      do_damage = True
      self.deal_damage(actorID, attackerID, 0, magic,
                       random.choice(['pois','fire','cold','elec','holy',
                                      'deth','tele','thef','vamp']))
    if 'thef' in dam_types:
      if not self.population[attackerID].inventory.full():
        if random.randint(1,100) <= 50 - self.calc_stat(actorID,'def'):
          choices = [i for i in actor.inventory.inv +
                     list(actor.equip.itervalues()) if i not in [0,-1,-2]]
          if choices != []:
            i = random.choice(choices)
            if self.try_place_item(i,(1,attackerID)):
              self.event_message(attackerID,i,'thef')
              #self.message(str(i)+self.items[i].name)
              if i in actor.equip:
                actor.equip.drop(i)
              else:
                actor.inventory.drop(i)
            else:
              self.message('stealing failed for no good reason')
          #self.message('thief!')
    if 'tele' in dam_types:
      if not self.resist(actorID,'tele'):
        do_damage = True
        if random.random() < 0.5:
          #self.event_message(actorID,0,'tele1')
          #self.message(actor.name+' disappears!')
          self.teleport(actorID)
          #self.event_message(actorID,0,'tele2')
          #self.message(actor.name+' appears!')
    if 'phys' in dam_types:
      if not self.resist(actorID, 'phys'):
        do_damage = True
    if 'magi' in dam_types:
      do_damage = True
    
    if do_damage:
      if 'vamp' in dam_types:
        self.event_message(attackerID,0,'vamp')
        self.heal(attackerID, min(actor.hp/6, magic/6))
      actor.hp -= max(0,damage - actor.armour())
    if actor.hp < 1:
      return True
    return False

  def actor_kill(self, actorID, attackerID):
    if actorID == 1 and debug:
      return
    
    self.event_message(actorID,0,'die')
    if actorID in self.allies:
      self.allies.remove(actorID)
      self.message(self.population[actorID].name+' is defeated!')
      if self.allies == []:
        self.message('You have won!!')
        self.population[1].status['WINNER'] = True
        self.uses['a tick of the clock'] = self.ticker.ticks
        self.message_uses()
      else:
        self.message('now go and kill '+' and '.join([
          self.population[enemyID].name for enemyID in self.allies])+'.')
    self.draw_messages()
    if actorID == 1:
      self.uses['a tick of the clock'] = self.ticker.ticks
      self.draw_status()
      self.message_uses()
      self.draw_messages()
      self.screen.getch()
    actor = self.population[actorID]
    floor = k.mget(self.grid,actor.location).items
    for itemID in actor.equip.itervalues():
      if itemID != 0:
        if not floor.full():
          floor.add(itemID)
        else:
          self.try_place_item(itemID,(0,self.random_container_location()))
    for itemID in actor.inventory.inv:
      if itemID not in [0,-1,-2]:
        if not floor.full():
          floor.add(itemID)
        else:
          self.try_place_item(itemID,(0,self.random_container_location()))
      
      
    self.population[attackerID].kill_count += 1
    self.population[attackerID].pray_count = max(
      0,self.population[attackerID].pray_count-1)
    self.del_actor(actorID)
  
  def actor_attack(self, attackerID, defenderID):
    attacker = self.population[attackerID]
    defender = self.population[defenderID]
    r = random.randint(1,100)
    death = False
    if random.randint(1,100) <= (80 + self.calc_stat(attackerID,'atk')
                                     - self.calc_stat(defenderID,'def')):
      self.event_message(attackerID,defenderID,'hit')
      m = 1
      if m > 1:
        print 'moo'
      if r <= (self.calc_stat(attackerID,'cat')
              - self.calc_stat(defenderID,'cdf')):
        m = 2
        while random.random() < 0.3:
          m += 1
      (roll, bonus) = self.calc_damage(attackerID)
      if m > 1 and roll > 0:
        self.event_message(attackerID,defenderID,m)
      death = self.deal_damage(defenderID, attackerID, (m*roll)+bonus,
                               m*roll, self.brands(attackerID))
      self.event_message(defenderID,0,defender.hp)
    else:
      self.event_message(attackerID,defenderID,'miss')
    if death:
      self.actor_kill(defenderID, attackerID)

  def move_actor(self, actorID, new_location):
    old_location = self.population[actorID].location
    if k.bound(old_location) == k.bound(new_location):
      return
    newtile = k.mget(self.grid, new_location)
    if newtile.blocked():
      if newtile.occupant != 0:
        self.actor_attack(actorID, newtile.occupant)
      return#raise TileBlockedException("tried to move actor to blocked tile")
    k.mget(self.grid, old_location).occupant = 0
    self.population[actorID].location = new_location
    newtile.occupant = actorID

  def teleport(self,actorID,regardless=False,messages=True):
    res = self.resist(actorID,'tele')
    if res and messages:
      self.event_message(actorID,0,'tele3')
    if regardless or not res:
      if messages:
        self.event_message(actorID,0,'tele1')
      self.move_actor(actorID, self.random_unblocked_location())
      if messages:
        self.event_message(actorID,0,'tele2')

  def move_item(self, itemID, (newloctype,newloc),(oldloctype,oldloc)):
    if debug:
      self.message('moving '+self.items[itemID].name)
      self.message(str((newloctype,newloc,oldloctype,oldloc)))
    if newloctype == 2:
      replacedID = self.population[newloc[0]].equip[newloc[1]]
    success = self.try_place_item(itemID, (newloctype, newloc))
    if success:
      if oldloctype == 3:
        pass
      elif oldloctype == 0:
        k.mget(self.grid,oldloc).items.drop(itemID)
      elif oldloctype == 1:
        self.population[oldloc].inventory.drop(itemID)
      elif oldloctype == 2:
        self.population[oldloc[0]].equip[oldloc[1]] = 0
    if newloctype == 2:
      if not self.try_place_item(replacedID,(1,newloc[0])):
        self.try_place_item(replacedID,(0,self.population[newloc[0]].location))
      #if not self.try_place_item(replacedID,(oldloctype,oldloc)):
      #  self.message('BAD BAD ERROR')
    return success

  def actor_walk(self, actorID, direction):
    self.move_actor(actorID, k.dirs[direction]
                    (self.population[actorID].location))

  def validate(self, actorID, d):
    if d == 5:
      return True
    ret = k.mget(self.grid, k.dirs[d]
                  (self.population[actorID].location)).walkable()
    return ret

  def tile_char(self, location):
    tile = k.mget(self.grid, location)
    if tile.occupant != 0:
      return (self.population[tile.occupant].char,
              self.population[tile.occupant].colour)
    if not tile.items.empty():
      return (self.items[tile.items.top()].char,
              self.items[tile.items.top()].colour)
    return tile.char(),curses.COLOR_WHITE

  def remember_tile_char(self,location):
    tile = k.mget(self.grid,location)
    howlong = self.ticker.ticks - tile.last_seen
    if howlong == 0:
      return self.tile_char(location)
    elif howlong < 200:
      return tile.player_memory, curses.COLOR_YELLOW
    else:
      return ' ', 0

  def brief_inv(self, inventory):
    return ''.join([self.items[i].char for i in inventory.inv])

  def verbose_inv(self, inventory):
    inv = []
    for n, itemID in sorted(inventory.iteritems()):
      if True:#itemID != 0:
        item = self.items[itemID]
        inv.append((str(n)+' '+item.char+' '+item.name)+self.empty_message
                   [:self.message_length])
    return inv

  def random_unblocked_location(self):
    for x in range(100):
      x,y = random.randint(0,width-1),random.randint(0,height-1)
      if not k.mget(self.grid, (x,y,False)).blocked():
        return (x,y,False)
    print "death"
    raw_input()

  def random_container_location(self):
    for x in range(100):
      x,y = random.randint(0,width-1),random.randint(0,height-1)
      if k.mget(self.grid, (x,y,False)).can_contain():
        return (x,y,False)
    print "death"
    raw_input()

  def get_dir(self):
    kp = self.screen.getch()
    if kp not in range(ord('1'),ord('9')+1):
      try:
        kp = k.vi_keys[kp]
      except KeyError:
        return (kp, False)
    else:
      kp -= ord('0')
    return (kp, True)

  def get_item_id2(self, inv1,inv2,message=True):
    itemID,slot = self.get_item_id2_aux(inv1,inv2)
    if itemID == 0:
      if message:
        self.message('Cancelled.')
    self.draw_colour(self.population[1].location)
    return itemID,slot
  
  def get_item_id2_aux(self, inv1,inv2):
    self.draw_inv2(inv1,inv2)
    toggle = False
    while True:
      kp = self.screen.getch()
      if kp - ord('0') in range(10):
        self.draw_colour(self.population[1].location)
        if toggle:
          return inv2[kp-ord('0')],kp-ord('0')
        else:
          return inv1[kp-ord('0')],kp-ord('0')
      elif kp == ord('-'):
        toggle = not toggle
        if toggle:
          self.draw_inv2(inv2,inv1)
        else:
          self.draw_inv2(inv1,inv2)
      else:
        return 0,0
    
  def get_item_id(self, inv,message=True):
    itemID,slot = self.get_item_id_aux(inv)
    if itemID == 0:
      if message:
        self.message('Cancelled.')
    self.draw_messages()
    return itemID,slot

  def get_item_id_aux(self, inv):
    self.draw_inv(inv)
    while True:
      kp = self.screen.getch()
      if kp - ord('0') in range(10):
        self.draw_colour(self.population[1].location)
        return inv[kp-ord('0')],kp-ord('0')
      else:
        return 0,0

  def get_inscription(self):
    self.screen.addstr(self.message_number,apx+2,'type inscription: ')
    inscr = ''
    c = self.screen.getch()
    i = 0
    while c in k.typable and i < 20:
      inscr = inscr + chr(c)
      self.screen.addch(chr(c))
      c = self.screen.getch()
      i += 1
    self.draw_messages()
    return inscr

  def actor_equip(self, actorID,itemID,messages=False):
    if itemID != 0:
      item = self.items[itemID]
      actor = self.population[actorID]
      oldhp = self.calc_stat(actorID,'hp')
      loc = actor.location
      if not item.equippable:
        if messages:
          self.message("Can't equip that!")
          self.draw_messages()
      else:
        replaceID = actor.equip[item.slot]
        if itemID not in actor.inventory:
          if self.move_item(itemID,(2,(actorID,item.slot)),(0,loc)):
            if messages:
              self.message('equipped ' + item.name + '.')
            newhp = self.calc_stat(actorID,'hp')
            actor.hp = min(newhp,max(1,actor.hp*newhp/oldhp))
            return True
        else:#in inventory
          if self.move_item(itemID,(2,(actorID,item.slot)),(1,actorID)):
            if messages:
              self.message('equipped ' + item.name + '.')
            newhp = self.calc_stat(actorID,'hp')
            actor.hp = min(newhp,max(1,actor.hp*newhp/oldhp))
            return True
    return False

  def los_blocked(self,loc):
    return k.mget(self.grid,loc).blocks_los()

  def update_tile(self, loc,ticks):
    tile = k.mget(self.grid,loc)
    tile.last_seen = ticks
    #tile.player_memory = self.tile_char(loc)
    tile.player_memory = tile.terrain

  def update_los(self, loc, ticks):
    k.tree_iter(fov_tree, lambda x:self.update_tile(x,ticks),
                self.los_blocked,loc)
    

  def player_turn(self, actorID):
    self.increment_uses('a turn')
    if actorID != 1:
      self.message("death")
      raw_input()
    actor = self.population[1]
    loc = self.population[1].location
    #self.message('')
    self.update_los(loc, self.ticker.ticks)
    #self.message(str(k.mget(self.grid,loc).last_seen)+str(self.ticker.ticks))
    self.draw_colour(self.population[1].location)
    while True:
      kp, is_direction = self.get_dir()
      if is_direction:#move
        if self.validate(1,kp):
          self.actor_walk(1,kp)
          return self.calc_stat(actorID,'spd')
      elif kp == ord('p'):#pick up/put down wall
        if not actor.loaded():
          self.message('Pick up the wall from which direction?')
          self.draw_messages()
          kp, is_direction = self.get_dir()
          if is_direction:
            tile = k.mget(self.grid,k.dirs[kp](loc))
            if tile.terrain != '.':
              if tile.terrain == '#':
                actor.inventory.add(-1)
              else:
                actor.inventory.add(-2)
              tile.terrain = '.'
              self.message('You pick up the wall.')
              return self.calc_stat(actorID,'spd')
            else:
              self.message('Cancelled.')
              self.draw_messages()
          else:
            self.message('Cancelled.')
            self.draw_messages()
        else: #actor.loaded()
          if -1 in actor.inventory or -2 in actor.inventory:
            self.message('Drop the wall in which direction?')
            self.draw_messages()
            kp, is_direction = self.get_dir()
            if is_direction:
              tile = k.mget(self.grid, k.dirs[kp](loc))
              if not tile.empty():
                self.message('Cancelled.')
                self.draw_messages()
              else:
                tile.terrain = actor.load()
                try:
                  actor.inventory.drop(-1)
                except kitems.InventoryException:
                  actor.inventory.drop(-2)
                self.message('You replace the wall.')
                return self.calc_stat(actorID,'spd')
          else:
            self.message('Carrying too much already!')
            self.draw_messages()
      elif kp == ord('w'):#wield
        itemID, slot = self.get_item_id2(actor.inventory,
                                         k.mget(self.grid,loc).items)
        if self.actor_equip(actorID,itemID,True):
          return self.calc_stat(actorID,'spd')
      elif kp == ord('e'):#equip
        self.message('equipment:')
        for slot, itemID in actor.equip.iteritems():
          if itemID != 0:
            self.message(kitems.indices[slot] + ': ' + self.items[itemID].name)
        self.draw_messages()
      elif kp == ord('d'):#drop to ground
        itemID, slot = self.get_item_id2(actor.inventory,actor.equip)
        if itemID == 0:
          self.message('Cancelled.')
        elif itemID in [-1,-2]:
          self.message('Use \'p\' instead.')
          self.draw_messages()
        elif not k.mget(self.grid,loc).can_contain():
          self.message('no space to put ' + self.items[itemID].name + '!')
          self.draw_messages()
        elif itemID in actor.inventory:
          if self.move_item(itemID,(0,loc),(1,1)):
            self.message('dropped ' + self.items[itemID].name + '.')
            self.draw_messages()
            self.draw_status()
        elif itemID in actor.equip:
          oldhp = self.calc_stat(1,'hp')
          if self.move_item(itemID,(0,loc),(2,(1,slot))):
            newhp = self.calc_stat(1,'hp')
            actor.hp = min(newhp,max(1,actor.hp*newhp/oldhp))
            self.message('dropped ' + self.items[itemID].name + '.')
            self.draw_messages()
            self.draw_status()
      elif kp == ord('g') or kp == ord(','):#get from ground
        if actor.inventory.full():
          self.message('inventory is full!')
          self.draw_messages()
        else:
          itemID,slot = self.get_item_id(k.mget(self.grid,loc).items)
          if itemID != 0:
            self.move_item(itemID,(1,1),(0,loc))
            self.message('got ' + self.items[itemID].name + '.')
            return self.calc_stat(actorID,'spd')
      elif kp == ord('Q'):#quit
        self.message('really quit? (y/n)')
        self.draw_messages()
        if self.screen.getch() == ord('y'):
          raise Exception('User quit')
        else:
          self.message('OK then.')
          self.draw_messages()
      elif kp == ord('t'):#take off
        itemID,slot = self.get_item_id(actor.equip)
        if itemID != 0:
          oldhp = self.calc_stat(1,'hp')
          if not actor.inventory.full():
            self.move_item(itemID,(1,1),(2,(1,slot)))
            self.message('took off ' + self.items[itemID].name + '.')
            newhp = self.calc_stat(1,'hp')
            actor.hp = min(newhp,max(1,actor.hp*newhp/oldhp))
            self.draw_messages()
            self.draw_status()
          elif not k.mget(self.grid,loc).items.full():
            self.move_item(itemID,(0,loc),(2,(1,slot)))
            self.message('took off and dropped ' +
                         self.items[itemID].name + '.')
            newhp = self.calc_stat(1,'hp')
            actor.hp = min(newhp,max(1,actor.hp*newhp/oldhp))
            self.draw_messages()
            self.draw_status()
          else:
            self.message('no space to put ' + self.items[itemID].name + '!')
      elif kp == ord('i'):#inventory/inspect
        while True:
          itemID,slot = self.get_item_id2(actor.inventory,actor.equip)
          if itemID != 0:
            self.draw_item(itemID)
            self.screen.getch()
          else:
            self.draw_messages()
            break
      elif kp == ord('q'):#quaff
        itemID,slot = self.get_item_id2(
          actor.inventory, k.mget(self.grid,loc).items,False)
        if itemID != 0 and isinstance(self.items[itemID],kitems.kpotion):
          if actor.inventory[slot] == itemID:
            self.quaff_potion_inventory(actorID,itemID)
          else:
            self.quaff_potion_floor(actorID,itemID)
          self.message('you quaff a '+self.items[itemID].name+'.')
          return self.calc_stat(actorID,'spd')
        else:
          self.message('Cancelled.')
          self.draw_messages()
      elif kp == ord('r'):#read
        itemID,slot = self.get_item_id2(
          actor.inventory, k.mget(self.grid,loc).items,False)
        if itemID != 0 and isinstance(self.items[itemID],kitems.kscroll):
          if actor.inventory[slot] == itemID:
            self.quaff_potion_inventory(actorID,itemID)
          else:
            self.quaff_potion_floor(actorID,itemID)
          self.message('you read a '+self.items[itemID].name+'.')
          return self.calc_stat(actorID,'spd')
        else:
          self.message('Cancelled.')
          self.draw_messages()
      elif kp == ord('x'):#examine
        xloc = loc
        fx,fy = 0,0#apx/2,1+apy/2
        self.draw_tile(k.mget(self.grid,xloc))
        self.screen.move(fy+1+apy/2,fx+apx/2+1)
        di = self.get_dir()
        while di[1]:
          fox,foy,u = k.naive_dirs[di[0]]((fx,fy,False))
          if abs(fox) < 1+apx/2 and abs(foy) < 1+apy/2:
            fx,fy = fox,foy
            xloc = k.dirs[di[0]](xloc)
            self.draw_tile(k.mget(self.grid,xloc))
            self.screen.move(fy+1+apy/2,fx+apx/2+1)
            di = self.get_dir()
          else:
            fox,foy = fx,fy
            di = self.get_dir()
        self.draw_messages()
      elif kp == ord('}'):
        itemID, slot = self.get_item_id2(actor.inventory,actor.equip)
        if itemID == 0:
          self.message('Cancelled.')
        else:
          self.items[itemID].inscription = ''
          self.items[itemID].update_name()
      elif kp == ord('{'):
        itemID,slot = self.get_item_id2(actor.inventory,actor.equip)
        if itemID == 0:
          self.message('Cancelled.')
        else:
          self.items[itemID].inscription = (self.items[itemID].inscription +
                                            self.get_inscription())
          self.items[itemID].update_name()
      elif kp == ord('P'):
        self.message('You pray.')
        self.increment_uses('prayer')
        actor.pray_count += 1
        actor.kill_count = max(0,actor.kill_count-1)
        #r = random.random()**(1+sqrt(actor.pray_count*actor.kill_count))
        r = max(0.0,random.random()-0.002*actor.pray_count)
        if r < 0.001:
          self.message('The Wrathful answers!')
          if self.deal_damage(1, 1, actor.hp/3, actor.hp/3,
                         ['cold','pois','fire','elec','deth','holy']):
            self.actor_kill(1,1)
          self.scr_rearrangement(1)
          self.scr_teleport(1)
          actor.pray_count = 0
        elif r < 0.5:
          self.message('No-one answers.')
        elif r < 0.65:
          self.message('The Lifegiver answers!')
          self.pot_healing(1)
        elif r < 0.8:
          self.message('The Earthshaper answers!')
          self.scr_rearrangement(1)
        elif r < 0.9:
          self.message('The Omnipresent answers!')
          self.scr_multiteleport(1,True)
        elif r < 0.95:
          self.message('The Omniscient answers!')
          for itemID in actor.inventory.inv:
            if itemID >0 and not self.items[itemID].equippable:
              self.items[itemID].identify()
        else:# r <= 1.0:
          self.message('The Trickster answers!!')
          self.draw_messages()
          for chumpID, chump in self.population.iteritems():
            if chumpID != 0 and chumpID not in self.daemons.values():
              oldhp = self.calc_stat(chumpID,'hp')
              for itemID in chump.inventory.inv:
                if itemID not in [0,-1,-2]:
                  chump.inventory.drop(itemID)
                  self.try_place_item(itemID,
                                      (0,self.random_container_location()))
              for itemID in chump.equip.eq.values():
                if itemID not in [0,-1,-2]:
                  chump.equip.drop(itemID)
                  self.try_place_item(itemID,
                                      (0,self.random_container_location()))
              newhp = self.calc_stat(chumpID,'hp')
              chump.hp = min(newhp,max(1,chump.hp*newhp/oldhp))
        return self.calc_stat(1,'spd')

        
  def greed_turn(self, actorID):
    actor = self.population[actorID]
    inventory = actor.inventory
    floor = k.mget(self.grid,actor.location).items
    for itemID in actor.inventory:
      if itemID != 0:
        item = self.items[itemID]
        if self.items[itemID].equippable:
          if self.items[actor.equip[item.slot]].value < item.value:
            self.actor_equip(actorID,itemID)
            self.event_message(actorID,0,'equip')
            return self.calc_stat(actorID,'spd')
    if floor.empty():
      return False
    for itemID in floor.inv:
      if itemID != 0:
        item = self.items[itemID]
        if item.equippable:
          if self.items[actor.equip[item.slot]].value < item.value:
            self.actor_equip(actorID,itemID)
            self.event_message(actorID,0,'equip')
            return self.calc_stat(actorID,'spd')
    return False
        
            
    
    

  def random_walk_turn(self, actorID):
    eq = self.greed_turn(actorID)
    if eq != False:
      return eq
    actor = self.population[actorID]
    loc = actor.location
    choices = [idx for idx, d in k.dirs.iteritems() if k.mget(self.grid,
                         d(loc)).walkable()]
    if choices != []:
      d = random.choice(choices)
      mem = actor.random_walk_memory
      if k.normalise(k.dirs[d](loc)) not in mem:
        self.actor_walk(actorID, d)
        actor.random_walk_memory = (mem + [k.normalise(k.dirs[d](loc))])[1:]
      else:#todo make this more efficient
        if random.choice([True] +[False]*(1+len(mem)-list(reversed(
          mem)).index(k.normalise(k.dirs[d](loc))))):
          return self.random_walk_turn(actorID)
        else:
          self.actor_walk(actorID, d)
          actor.random_walk_memory = (actor.random_walk_memory
                                    + [k.normalise(k.dirs[d](loc))])[1:]
    return self.calc_stat(actorID,'spd')

  def digger_turn(self, actorID):
    eq = self.greed_turn(actorID)
    if eq != False:
      return eq
    actor = self.population[actorID]
    actor.last_pickup += 1
    if random.random() < actor.pickup_prob():
      if not actor.loaded():
        choices = [idx for idx, d in k.dirs.iteritems() if
                   k.mget(self.grid,d(actor.location)).occupant == 0]
        if choices == []:
          return self.random_walk_turn(actorID)
        di = random.choice(choices)
        tile = k.mget(self.grid,k.dirs[di](actor.location))
        if tile.terrain != '.':
          if tile.terrain == '#':
            actor.inventory.add(-1)
          else:
            actor.inventory.add(-2)
          tile.terrain = '.'
          actor.last_pickup = 0
        else:
          self.actor_walk(actorID,di)
        return self.calc_stat(actorID,'spd')
      else: #actor.loaded()
        choices = [idx for idx,d in k.dirs.iteritems() if
                   k.mget(self.grid,d(actor.location)).empty()]
        if choices == [] or (-1 not in actor.inventory) or (
              -2 not in actor.inventory):
          return self.random_walk_turn(actorID)
        di = random.choice(choices)
        tile = k.mget(self.grid,k.dirs[di](actor.location))
        tile.terrain = actor.load()
        try:
          actor.inventory.drop(-1)
        except kitems.InventoryException:
          actor.inventory.drop(-2)
        actor.last_pickup = 0
      return self.calc_stat(actorID,'spd')
    else:
      return self.random_walk_turn(actorID)

  def scent_chase_turn(self, actorID): #todo account for vision, sound
    eq = self.greed_turn(actorID)
    if eq != False:
      return eq
    actor = self.population[actorID]
    loc = actor.location
    sc = self.scent_maps[actor.target]
    for idx,d in k.dirs.iteritems():
      if k.mget(self.grid, d(loc)).occupant == actor.target:
        self.actor_walk(actorID,idx)
        return self.calc_stat(actorID,'spd')
    choices = [(k.mget(sc,d(loc)),idx) for idx,d in k.dirs.iteritems()
               if k.mget(self.grid,d(loc)).walkable() and idx != 5
               and k.mget(self.grid,d(loc)).occupant not in actor.allies]
    if choices != []:
      self.actor_walk(actorID, max(choices)[1])
    return self.calc_stat(actorID,'spd')

  def scent_flee_turn(self, actorID):
    eq = self.greed_turn(actorID)
    if eq != False:
      return eq
    actor = self.population[actorID]
    loc = actor.location
    sc = self.scent_maps[actor.target]
    choices = [(k.mget(sc,d(loc)),idx) for idx,d in k.dirs.iteritems()
               if k.mget(self.grid,d(loc)).walkable() and idx != 5
               and k.mget(self.grid,d(loc)).occupant not in actor.allies]
    if choices != []:
      self.actor_walk(actorID, min(choices)[1])
    return self.calc_stat(actorID,'spd')
    
  def add_scent_map(self, actorID):
    if actorID not in self.scent_maps:
      self.scent_maps[actorID] = [[0.0]*width for y in range(height)]
      k.mset(self.scent_maps[actorID],self.population[actorID].location,1.0)

  
  def update_scent_map(self, scent_map, t_loc):
    new_map = [[0.0]*width for h in range(height)]
    for x in range(width):
      for y in range(height):
        if not k.mget(self.grid,(x,y,False)).walkable():
          new_map[y][x] = 0.0
        else:
          neighbours = [d((x,y,False)) for idx,d in k.dirs.iteritems() #incl idx==5
                        if k.mget(self.grid, d((x,y,False))).walkable()]
          if (x,y,False) == k.normalise(t_loc):
          ## HACK: pass a normalised loc if actor is alive,
          ## else pass (0,0,True) and fail this test
          ## atm only the player has scent #ignore this
            if neighbours == []:
              new_map[y][x] = 5.0+0.99*(k.mget(scent_map,(x,y,False)))
            else:
              new_map[y][x] = 5.0+0.99*(sum([k.mget(scent_map,loc)
                            for loc in neighbours]))/len(neighbours)
          else:
            if neighbours == []:
              new_map[y][x] = 0.99*k.mget(scent_map,(x,y,False))
            else:
              new_map[y][x] = 0.99*sum([k.mget(scent_map,loc)
                              for loc in neighbours])/len(neighbours)
    return new_map

  def pot_healing(self,actorID):
    actor = self.population[actorID]
    max_hp = self.calc_stat(actorID,'hp')
    actor.hp = min(max_hp, (actor.hp + max_hp)/2)
    self.message('you feel better.')
    self.draw_status()

  def pot_poison(self,actorID):
    if not self.resist(actorID, 'pois'):
      self.inflict(actorID, 'pois', self.population[actorID].hp/2,1)
      self.event_message(actorID,0,'pois1')
      #self.message('you feel very sick...')
    else:
      self.message('that tastes disgusting!')
      
  def pot_speed(self,actorID):
    self.inflict(actorID,'fast', 100, 0)
    self.event_message(actorID,0,'fast1')
    #self.message('you feel very fast!')

  def pot_ice(self,actorID):
    if not self.resist(actorID, 'cold'):
      self.inflict(actorID, 'slow', 100,1)
      self.event_message(actorID,0,'cold1')
      #self.message('you feel very sick...')
    else:
      self.message('Refreshingly cool!')

  def pot_resistance(self,actorID):
    self.inflict(actorID,'resi',100,0)
    self.event_message(actorID,0,'res1')

  def pot_berserkergang(self,actorID):
    self.inflict(actorID,'berz',100,0)
    self.event_message(actorID,0,'berz1')

  def pot_skill(self,actorID):
    self.event_message(actorID,0,'skill')
    actor = self.population[actorID]
    stats = actor.base_stats
    oldhp = self.calc_stat(1,'hp')
    for i in range(3):
      stat = random.choice(stats.keys())
      stats[stat] += 2
      if stat == 'hp':
        stats[stat] += 3
        newhp = self.calc_stat(1,'hp')
        actor.hp = min(newhp,max(1,actor.hp*newhp/oldhp))

  def scr_teleport(self,actorID,regardless=False):
    self.teleport(actorID,regardless)
    #self.message('you disappear!')

  def scr_multiteleport(self,actorID,regardless=False):
    self.message('everyone disappears!')
    for aID,actor in self.population.iteritems():
      if not isinstance(actor,kactors.kdaemon) and aID != 0:
        self.teleport(aID,regardless=regardless,messages=False)
  
  def scr_rearrangement(self,actorID):
    self.message('The ground churns and reforms!!')
    epicentre = self.population[actorID].location
    radius = 2
    displacement = (0,0)
    loc = epicentre
    prev = epicentre
    for i in range(15):
      for j in range(2):
        di,displacement = random.choice(k.rearr_step(displacement,radius))
        loc = k.dirs[di](loc)
      prev = loc
      for j in range(3):
        di,displacement = random.choice(k.rearr_step(displacement,radius))
        loc = k.dirs[di](loc)
      self.rearr_swap(prev,loc)

  def rearr_swap(self,loc1,loc2):
    if loc1 == loc2:
      return
    tile1 = k.mget(self.grid,loc1)
    tile2 = k.mget(self.grid,loc2)
    terrain1,inv1,occ1 = tile1.terrain,tile1.items,tile1.occupant
    terrain2,inv2,occ2 = tile2.terrain,tile2.items,tile2.occupant
    if occ1 != 0:
      self.population[occ1].location = loc2
    if occ2 != 0:
      self.population[occ2].location = loc1
    tile1.terrain = terrain2
    tile2.terrain = terrain1
    tile1.occupant = occ2
    tile2.occupant = occ1
    tile1.items = inv2
    tile2.items = inv1

  def scr_loss(self,actorID):
    inv = self.population[actorID].inventory
    eq = self.population[actorID].equip
    choices = [i for i in inv.inv+list(eq.itervalues()) if i not in [0,-1,-2]]
    if choices != []:
      itemID = random.choice(choices)
      self.message(self.items[itemID].name + ' vanishes!')
      if self.try_place_item(itemID,(0,self.random_container_location())):
        if itemID in eq:
          eq.drop(itemID)
        else:
          inv.drop(itemID)
    else:
      self.message('nothing happens.')

  def scr_enchantment(self,actorID):
    inv = self.population[actorID].inventory
    eq = self.population[actorID].equip
    choices = [i for i in inv.inv+list(eq.itervalues()) if i not in [0,-1,-2]
               and self.items[i].equippable]
    if choices == []:
      self.message('nothing happens.')
    else:
      item = self.items[random.choice(choices)]
      item.enchant()
      self.message(item.name+' glows brightly!')

  def scr_damage(self,actorID):
    inv = self.population[actorID].inventory
    eq = self.population[actorID].equip
    choices = [i for i in inv.inv+list(eq.itervalues()) if i not in [0,-1,-2]
               and self.items[i].equippable]
    if choices == []:
      self.message('nothing happens.')
    else:
      actor = self.population[actorID]
      oldhp = self.calc_stat(actorID,'hp')
      item = self.items[random.choice(choices)]
      item.damage()
      newhp = self.calc_stat(actorID,'hp')
      actor.hp = min(newhp,max(actorID,actor.hp*newhp/oldhp))
      self.message(item.name+' decays!')

  def scr_summoning(self,actorID):
    a = [aid for aid in self.allies if self.population[aid].char != 'p']
    if a == []:
      self.message('nothing happens.')
      return
    aid = random.choice(a)
    actor = self.population[actorID]
    choices = [idx for idx, d in k.dirs.iteritems() if k.mget(self.grid,
                         d(actor.location)).walkable() and idx != 5]
    if choices == []:
      self.message('nothing happens.')
      return
    self.message(self.population[aid].name+' is summoned to you!')
    self.move_actor(aid,k.dirs[random.choice(choices)](actor.location))

  def quaff_potion_inventory(self,actorID,itemID):#also for scrolls
    self.population[actorID].inventory.drop(itemID)
    self.items[itemID].function(actorID)
    self.items[itemID].identify()
    self.increment_uses(self.items[itemID].name)
  
  def quaff_potion_floor(self,actorID,itemID):#also for scrolls
    k.mget(self.grid,self.population[actorID].location).items.drop(itemID)
    self.items[itemID].function(actorID)
    self.items[itemID].identify()
    self.increment_uses(self.items[itemID].name)

  def start_game(self):
    self.message('welcome to KleinRL! Did you rtfm.txt?')
    self.message('You will be chased by '+str(enemies)+' villains today.')
    self.add_actor(kactors.kplayer(self.player_turn),
                   self.random_unblocked_location())
    self.add_scent_map(1)
    for func, value in [(self.pot_healing,100),(self.pot_poison,0),
                        (self.pot_speed,2000),(self.pot_berserkergang,9000),
                        (self.pot_resistance,1000),(self.pot_ice,0),
                        (self.pot_skill,10000)]:
      self.potions.append(self.add_item(kitems.kpotion(func,value),(3,0)))
    for func,value in [(self.scr_teleport,100),(self.scr_multiteleport,500),
                       (self.scr_rearrangement,1000),(self.scr_loss,0),
                       (self.scr_enchantment,10000),(self.scr_damage,0),
                       (self.scr_summoning,0)]:
      self.scrolls.append(self.add_item(kitems.kscroll(func,value),(3,0)))
    
    p_loc = self.population[1].location
    if debug:
      self.add_item(kitems.random_weapon(),(0,p_loc))
      self.add_item(kitems.random_armour(),(0,p_loc))
      self.add_item(kitems.random_boots(),(0,p_loc))
      self.add_item(kitems.random_helmet(),(0,p_loc))
      self.add_item(kitems.random_shield(),(0,p_loc))
      self.add_item(kitems.random_gloves(),(0,p_loc))
      self.add_item(kitems.random_ring(),(0,p_loc))
    self.add_item(kitems.random_weapon(),(0,p_loc))
    eqchoices = [kitems.random_weapon,kitems.random_armour,
                     kitems.random_helmet,kitems.random_shield,
                     kitems.random_gloves,kitems.random_ring,
                     kitems.random_amulet,kitems.random_boots]
    self.add_item(kitems.random_weapon(),(0,p_loc))
    #self.add_item(random.choice(eqtypes)(),(0,p_loc))
    for i in range(40):
      itype = random.choice(eqchoices)
      self.add_item(itype(),(0,self.random_container_location()))
    for i in range(15):
      self.try_place_item(random.choice(self.scrolls),
                          (0,self.random_container_location()))
      self.try_place_item(random.choice(self.potions),
                          (0,self.random_container_location()))


    #daemons
    self.daemons['pois'] = self.add_daemon(kactors.kdaemon(
      self.poison_damage, 'poison'))
    self.daemons['fire'] = self.add_daemon(kactors.kdaemon(
      self.fire_damage, 'fire'))
    self.daemons['slow'] = self.add_daemon(kactors.kdaemon(
      self.slow, 'slowness'))
    self.daemons['conf'] = self.add_daemon(kactors.kdaemon(
      self.conf, 'confusion'))
    self.daemons['fast'] = self.add_daemon(kactors.kdaemon(
      self.fast, 'quickness'))
    self.daemons['resi'] = self.add_daemon(kactors.kdaemon(
      self.temp_resi,'resistance'))
    self.daemons['berz'] = self.add_daemon(kactors.kdaemon(
      self.berserk,'berserkergang'))

    #zombies
    for i in range(6):
      aid = self.add_actor(kactors.kzombie(self.random_walk_turn),
                     self.random_unblocked_location())
      itype = random.choice(eqchoices)
      self.add_item(itype(),(0,self.population[aid].location))
      itype = random.choice(eqchoices)
      self.add_item(itype(),(0,self.population[aid].location))
    self.allies = []

    if enemies != 0:
      enemy_dict = kdata.enemies
      for enemy in range(enemies):
        letter = random.choice(enemy_dict.keys())
        name = random.choice(enemy_dict[letter])
        del enemy_dict[letter]
        aid = self.add_actor(kactors.kactor(letter,self.scent_chase_turn,
                             name), self.random_unblocked_location())
        self.allies.append(aid)
        self.population[aid].target = 1
        self.add_item(kitems.random_weapon(),(0,self.population[aid].location))
        for i in range(int(sqrt(enemies))):
          itype = random.choice(eqchoices)
          self.add_item(itype(),(0,self.population[aid].location))
    
    #the puppetmaster
    aid = self.add_actor(kactors.kactor('p',self.scent_flee_turn,
                                        'The Puppetmaster'),
                   self.random_unblocked_location())
    self.allies.append(aid)
    itype = random.choice(eqchoices[1:])
    self.add_item(itype(),(0,self.population[aid].location))
    itype = random.choice(eqchoices[1:])
    self.add_item(itype(),(0,self.population[aid].location))
    itype = random.choice(eqchoices[1:])
    self.add_item(itype(),(0,self.population[aid].location))
    #self.add_scent_map(aid)
    self.population[aid].target = 1

    for ally in self.allies:
      self.population[ally].allies = self.allies

    #diggers
    for i in range(5):
      aid = self.add_actor(kactors.kdigger(self.digger_turn),
                     self.random_unblocked_location())
      itype = random.choice(eqchoices)
      self.add_item(itype(),(0,self.population[aid].location))

    for actorID, actor in self.population.iteritems():
      if actorID != 0:
        self.ticker.schedule_turn(self.calc_stat(actorID,'spd'), actorID)

  def game_turn(self):
    if random.random() < 0.02:
      itype = random.choice([kitems.random_weapon,kitems.random_armour,
                     kitems.random_helmet,kitems.random_shield,
                     kitems.random_gloves,kitems.random_ring,
                     kitems.random_amulet,kitems.random_boots])
      self.add_item(itype(),(0,self.random_container_location()))
    if random.random() < 0.04:
      self.try_place_item(random.choice(self.scrolls),
                          (0,self.random_container_location()))
      self.try_place_item(random.choice(self.potions),
                          (0,self.random_container_location()))
    
    if random.random()<0.01:
      aid = self.add_actor(kactors.kdigger(self.digger_turn),
                     self.random_unblocked_location())
      self.ticker.schedule_turn(self.calc_stat(aid,'spd'), aid)
    if random.random() < 0.02:
      aid = self.add_actor(kactors.kzombie(self.random_walk_turn),
                     self.random_unblocked_location())
      self.ticker.schedule_turn(self.calc_stat(aid,'spd'), aid)
      
    removal = []
    for actorID in self.scent_maps:
      if actorID in self.population:
        new_scent_map = self.update_scent_map(
          self.scent_maps[actorID], k.normalise(
            self.population[actorID].location))
        self.scent_maps[actorID] = new_scent_map
      else:
        removal.append(actorID)
    for actorID in removal:
      del self.scent_maps[actorID]
    for actorID in self.ticker.next_turn():
      if actorID in self.population:
        interval = self.population[actorID].do_turn(actorID)
        self.ticker.schedule_turn(interval, actorID)
      elif actorID == 1:
        return False
    return True

  def draw(self, loc):
    line_no = 1
    for line in k.sized_viewport(loc, apx, apy):
      self.screen.addstr(line_no, 1, ''.join(map(self.remember_tile_char,line)))
      line_no += 1
    self.draw_status()
    self.draw_messages()#includes refresh

  def draw_colour(self, loc):
    line_no = 1
    for line in k.sized_viewport(loc,apx,apy):
      self.screen.move(line_no,1)
      for char,colour in map(self.remember_tile_char,line):
        self.screen.addch(char,curses.color_pair(colour))
      line_no += 1
    self.draw_status()
    self.draw_messages()
    
  def draw_messages(self):
    line_no = 1
    for message in self.messages[-self.message_number:]:
      self.screen.addstr(line_no, apx+2, k.capitalise(message))
      line_no += 1
    self.screen.refresh()

  def draw_status(self):
    player = self.population[1]
    self.screen.addstr(apy+1,1,self.brief_inv(
      self.population[1].inventory)+'inven')
    self.screen.addstr(apy+2,1,''.join(map(str,range(10)))+' keys')
    self.screen.addstr(apy+3,1,self.brief_inv(
      k.mget(self.grid,self.population[1].location).items)+'floor')
    hpdisp = 'HP: '+str(player.hp)+'/'+str(self.calc_stat(1,'hp'))
    effects = (' '.join(player.status.keys()))
    self.screen.addstr(apy+4,1,(hpdisp+' '*apx)[:apx])
    spddisp = 'speed: '+ str(self.calc_stat(1,'spd'))
    self.screen.addstr(apy+5,1,(spddisp+' '*apx)[:apx])
    self.screen.addstr(self.message_number+1,1,(effects+
                  ' '*(apx+self.message_length))[:apx+self.message_length])
##    self.screen.addstr(self.message_number+1,apx+1,self.brief_inv(
##      self.population[1].inventory)+'inven')
##    self.screen.addstr(self.message_number+2,apx+1,
##                       ''.join(map(str,range(10)))+' keys')
##    self.screen.addstr(self.message_number+3,apx+1,self.brief_inv(
##      k.mget(self.grid,self.population[1].location).items)+'floor')
##    self.screen.addstr(self.message_number+4,apx+1,
##                ('HP: '+str(player.hp) + '/'+str(self.calc_stat(1,'hp'))+
##                ' speed: '+str(self.calc_stat(1,'spd'))+
##                             self.empty_message)[:self.message_length])
##    self.screen.addstr(self.message_number+5,apx+1,self.empty_message)
##    self.screen.addstr(self.message_number+5,apx+1,
##                       ' '.join(player.status.keys()))
  
  def draw_inv(self, inv):
    line_no = 1
    for line in self.verbose_inv(inv):
      self.screen.addstr(line_no,apx+2,line[:self.message_length])
      line_no += 1
    self.screen.refresh()

  def draw_tile(self,tile):
    if tile.last_seen < self.ticker.ticks:
      self.draw_messages()
      self.screen.addstr(1,apx+2,('tile out of sight'+self.empty_message
                          )[:self.message_length])
      return
    if tile.occupant != 0:
      self.screen.addstr(1,apx+2,(tile.char()+' '+
                                  self.population[tile.occupant].name +
                                  self.empty_message)[:self.message_length])
    else:
      self.screen.addstr(1,apx+2,(tile.char()+' no-one'+self.empty_message
                                  )[:self.message_length])
    line_no = 2
    for line in self.verbose_inv(tile.items):
      self.screen.addstr(line_no,apx+2,line[:self.message_length])
      line_no += 1
    self.screen.refresh()
  
  def draw_item(self, itemID):
    item = self.items[itemID]
    if item.equippable:
      self.message(item.char + ' ' + item.name + ', ' + item.slot)
      if debug:
        self.message(str(' '.join(item.stats.keys())))
        self.message(str('   '.join(map(str,item.stats.values()))))
        self.message(str(' '.join(list(item.brands))))
        self.message(str(' '.join(list(item.resistances))))
    else:
      self.message(item.char + ' ' +item.name+', ')
    if debug:
      self.message(str(item.value))
    self.draw_messages()
  
  def draw_inv2(self, inv1,inv2):
    line_no = 1
    for line in self.verbose_inv(inv1):
      self.screen.addstr(line_no,apx+2,line[:self.message_length])
      line_no += 1
    for line in self.verbose_inv(inv2):
      self.screen.addstr(line_no,apx+2,'-'+line[:self.message_length-1])
      line_no += 1
    self.screen.refresh()

  def message_uses(self):
    self.message('you used the following:')
    count = 1
    for name, count in self.uses.iteritems():
      self.message(name+', '+str(count)+' times,')
      count += 1
      if count == 18:
        self.message('--more--')
        self.draw_messages()
        self.screen.getch()
        count = 1
    self.message('and that\'s it.')
    self.message('You were chased by '+str(enemies)+' villains.')
  def increment_uses(self, name):
    if self.uses.has_key(name):
      self.uses[name] += 1
    else:
      self.uses[name] = 1

  def message(self, msg,important=True):
    m = (msg+self.empty_message)[0:self.message_length]
    if not important:
      if m not in self.messages[-3:]:
        self.messages.append(m)
    else:
      self.messages.append(m)

  def main(self):
    self.start_game()
    while self.game_turn():
      pass

def main(stdscr):
  for colour in range(1,8):
    curses.init_pair(colour,colour,0)
  stdscr.addstr(0,0,"How many enemies would you like to chase you today?")
  stdscr.addstr(1,0,"If you haven't played before, "+
                "perhaps you'd better practice with 0.")
  stdscr.addstr(2,0,"Once you've learnt the ropes, I recommend playing with 2.")
  stdscr.addstr(3,0,"If you want more of a challenge, "+
                "you can try any number up to 20.")
  s = ''
  while True:
    kp = stdscr.getch()
    if kp in range(ord('0'),ord('9')+1):
      s = s + chr(kp)
      stdscr.addstr(4,4,s)
    elif len(s) > 0:
      #stdscr.addstr(0,0,str(('0'+s,int('0'+s))))
      #stdscr.getch()
      break
  global enemies
  enemies = min(20,abs(int('0'+s)))
  stdscr.clear()
  game = krl(stdscr)
  game.main()

if __name__ == "__main__":
  try:
    curses.wrapper(main)
  except curses.error:
    print 'KleinRL needs at least an 72*22 terminal to run.'
